@@ -5,15 +5,20 @@ import platform
 import re
 import signal
 import traceback
-from sexpdata import dumps, loads, Symbol, Quoted
+import time
+from sexpdata import dumps, loads, parse
+from sexpdata import Symbol, Quoted
 from threading import Event
 
 from euslime.bridge import AbortEvaluation
-from euslime.bridge import EuslispError, EuslispInternalError, EuslispFatalError
+from euslime.bridge import EuslispError
+from euslime.bridge import EuslispFatalError
+from euslime.bridge import EuslispInternalError
 from euslime.bridge import EuslispProcess
 from euslime.bridge import EuslispResult
 from euslime.bridge import no_color
 from euslime.logger import get_logger
+from euslime.bridge import DELIM
 
 log = get_logger(__name__)
 REGEX_LISP_VECTOR = re.compile(r' \\(#[0-9]*[a-zA-Z]*) \(')
@@ -46,19 +51,31 @@ def qstr(s):
     # double escape characters for string formatting
     return s.encode('utf-8').replace('"', '\\"')
 
+
 def unquote(s):
     if isinstance(s, Quoted):
         # For instance in emacs 27
         return s.value()
     return s[1]  # [Symbol('quote'), [ ... ]]
 
+
 def dumps_lisp(s):
     return dumps(s, true_as='lisp:t', false_as='lisp:nil', none_as='lisp:nil')
 
+
 def dumps_vec(s):
     # Fix vector notation, which is not well supported by sexpdata
     return REGEX_LISP_VECTOR.sub(r' \1(', dumps(s))
 
+
+def check_lock(lock, timeout):
+    # lock.acquire(timeout=timeout) available from python3
+    start = time.time()
+    while lock.locked() and (time.time() - start) < timeout:
+        time.sleep(0.01)
+    return not lock.locked()
+
+
 class DebuggerHandler(object):
     restarts = [
         ["QUIT", "Quit to the SLIME top level"],
@@ -100,6 +117,15 @@ class DebuggerHandler(object):
         ]
         return res
 
+    def make_debug_return(self, response_type='abort', delim=DELIM):
+        if self.message:
+            msg = self.message.split(delim)[0]
+            msg = repr(msg.rsplit(' in ', 1)[0])
+        else:
+            msg = None
+        yield [Symbol(':debug-return'), 0, self.level, Symbol('nil')]
+        yield [Symbol(':return'), {response_type: msg}, self.id]
+
 
 class EuslimeHandler(object):
     def __init__(self, *args, **kwargs):
@@ -115,13 +141,18 @@ class EuslimeHandler(object):
         on_output = self.euslisp.on_output
         color = self.euslisp.color
         self.euslisp.stop()
-        self.euslisp = EuslispProcess(program, init_file, on_output=on_output, color=color)
+        self.euslisp = EuslispProcess(program,
+                                      init_file,
+                                      on_output=on_output,
+                                      color=color)
         self.euslisp.start()
-        self.euslisp.reset()  # get rid of the first abort request on the reploop
+        # get rid of the first abort request on the reploop
+        self.euslisp.reset()
         self.euslisp.accumulate_output = False
 
     def maybe_new_prompt(self):
-        new_prompt = self.euslisp.exec_internal('(slime::slime-prompt)', force_repl_socket=True)
+        cmd = '(slime::slime-prompt)'
+        new_prompt = self.euslisp.exec_internal(cmd, force_repl_socket=True)
         if new_prompt:
             yield [Symbol(":new-package")] + new_prompt
 
@@ -139,25 +170,48 @@ class EuslimeHandler(object):
             return [dumps_vec(result), True]
         return None
 
+    def eval_repl_result(self, cmd):
+        for val in self.euslisp.eval(cmd):
+            if isinstance(val, str):
+                # Colors are not allowed in :repl-result formatting
+                yield [Symbol(":write-string"),
+                       no_color(val),
+                       Symbol(":repl-result")]
+                yield [Symbol(":write-string"),
+                       "\n",
+                       Symbol(":repl-result")]
+            else:
+                yield val
+
     def _emacs_return_string(self, process, count, msg):
         self.euslisp.input(msg)
+        if self.euslisp.read_busy:
+            log.debug("Exitting read mode...")
+            self.euslisp.read_busy = False
         if len(msg) % 128 == 0:
             # communicate when the message ends exactly at buffer end
-            self.euslisp.exec_internal('(send slime::*slime-input-stream* :set-flag)')
+            cmd = '(send slime::*slime-input-stream* :set-flag)'
+            self.euslisp.exec_internal(cmd)
         if self.euslisp.read_mode:
-            yield [Symbol(":read-string"), 0, 1]
+            yield [Symbol(":read-string"), process, 1]
 
     def _emacs_interrupt(self, process):
         if self.euslisp.read_mode:
+            # propagate the signal to all child processes
             os.kill(- os.getpgid(self.euslisp.process.pid), 2)
         else:
+            # send the signal directly to the euslisp process
             self.euslisp.process.send_signal(signal.SIGINT)
-        if isinstance(process,int):
+        if isinstance(process, int):
+            # during a :read-string call
+            self.euslisp.read_busy = False
             yield [Symbol(":read-aborted"), process, 1]
 
     def swank_connection_info(self):
-        version = self.euslisp.exec_internal('(slime::implementation-version)')
-        name = self.euslisp.exec_internal('(lisp:pathname-name lisp:*program-name*)')
+        cmd_version = '(slime::implementation-version)'
+        cmd_name = '(lisp:pathname-name lisp:*program-name*)'
+        version = self.euslisp.exec_internal(cmd_version)
+        name = self.euslisp.exec_internal(cmd_name)
         res = {
             'pid': os.getpid(),
             'style': False,
@@ -183,9 +237,17 @@ class EuslimeHandler(object):
         yield EuslispResult(res)
 
     def swank_create_repl(self, sexp):
-        res = self.euslisp.exec_internal('(slime::slime-prompt)', force_repl_socket=True)
-        self.euslisp.accumulate_output = False
-        yield EuslispResult(res)
+        lock = self.euslisp.euslime_connection_lock
+        log.debug('Acquiring lock: %s' % lock)
+        lock.acquire()
+        cmd = '(slime::slime-prompt)'
+        try:
+            res = self.euslisp.exec_internal(cmd, force_repl_socket=True)
+            self.euslisp.accumulate_output = False
+            yield EuslispResult(res)
+        finally:
+            if lock.locked():
+                lock.release()
 
     def swank_repl_create_repl(self, *sexp):
         return self.swank_create_repl(sexp)
@@ -198,15 +260,12 @@ class EuslimeHandler(object):
         log.debug('Acquiring lock: %s' % lock)
         lock.acquire()
         try:
-            for val in self.euslisp.eval(sexp):
-                if isinstance(val,str):
-                    # Colors are not allowed in :repl-result formatting
-                    yield [Symbol(":write-string"), no_color(val), Symbol(":repl-result")]
-                    yield [Symbol(":write-string"), "\n", Symbol(":repl-result")]
-                else:
-                    yield val
-            if self.euslisp.read_mode:
+            for val in self.eval_repl_result(sexp):
+                yield val
+            if self.euslisp.read_mode or self.euslisp.read_busy:
+                log.debug("Aborting read mode...")
                 self.euslisp.read_mode = False
+                self.euslisp.read_busy = False
                 yield [Symbol(":read-aborted"), 0, 1]
             for val in self.maybe_new_prompt():
                 yield val
@@ -214,8 +273,10 @@ class EuslimeHandler(object):
             lock.release()
         except AbortEvaluation as e:
             log.info('Aborting evaluation...')
-            if self.euslisp.read_mode:
+            if self.euslisp.read_mode or self.euslisp.read_busy:
+                log.debug("Aborting read mode...")
                 self.euslisp.read_mode = False
+                self.euslisp.read_busy = False
                 yield [Symbol(":read-aborted"), 0, 1]
             for val in self.maybe_new_prompt():
                 yield val
@@ -227,8 +288,10 @@ class EuslimeHandler(object):
                 yield EuslispResult(None)
         except Exception as e:
             log.error(traceback.format_exc())
-            if self.euslisp.read_mode:
+            if self.euslisp.read_mode or self.euslisp.read_busy:
+                log.debug("Aborting read mode...")
                 self.euslisp.read_mode = False
+                self.euslisp.read_busy = False
                 yield [Symbol(":read-aborted"), 0, 1]
             if lock.locked():
                 lock.release()
@@ -243,6 +306,18 @@ class EuslimeHandler(object):
     def swank_repl_listener_eval(self, sexp):
         if self.debugger:
             return self.swank_throw_to_toplevel()
+        # quick fix for dealing with multiple s-expressions (#12)
+        try:
+            pexp = parse(sexp)
+            # anything that starts with an sexp and has some trailing part
+            if len(pexp) > 1 and type(pexp[0]) == list:
+                log.debug("Multiple s-expressions detected! Wrapping in progn")
+                sexp = "(progn {})".format(sexp)
+        except Exception:
+            # catch any possible parsing exceptions;
+            # we want to evaluate it anyways
+            log.error(traceback.format_exc())
+
         return self.swank_eval(sexp)
 
     def swank_pprint_eval(self, sexp):
@@ -262,6 +337,12 @@ class EuslimeHandler(object):
      (prompt quicklisp-client:*quickload-prompt*) explain &allow-other-keys)"
    t))
  19)
+
+  from swank-arglists.lisp:
+Return a list of two elements.
+First, a string representing the arglist for the deepest subform in
+RAW-FORM that does have an arglist.
+Second, a boolean value telling whether the returned string can be cached.
         """
         try:
             sexp = unquote(sexp)
@@ -297,14 +378,18 @@ class EuslimeHandler(object):
     def swank_simple_completions(self, start, pkg):
         # (swank:simple-completions "vector-" (quote "USER"))
         pkg = unquote(pkg)
-        cmd = '(slime::slime-find-symbol "{0}" "{1}")'.format(qstr(start), qstr(pkg))
+        cmd = '(slime::slime-find-symbol "{0}" "{1}")'.format(
+            qstr(start), qstr(pkg))
         yield EuslispResult(self.euslisp.exec_internal(cmd))
 
     def swank_fuzzy_completions(self, start, pkg, *args):
-        # Unsupported. Just returning a list of simple_completions in the fuzzy format
-        simple_completions = self.swank_simple_completions(start,pkg).next().value
+        # Unsupported
+        # Return a list of simple_completions in the fuzzy format
+        gen = self.swank_simple_completions(start, pkg)
+        simple_completions = gen.next().value
         if simple_completions:
-            simple_completions = [[x, 0, None, None] for x in simple_completions[0]]
+            first_comp = simple_completions[0]
+            simple_completions = [[x, 0, None, None] for x in first_comp]
         yield EuslispResult([simple_completions, None])
 
     def swank_fuzzy_completion_selected(self, original_string, completion):
@@ -318,14 +403,32 @@ class EuslimeHandler(object):
                 scope = scope[:-1]  # remove marker
                 if scope[-1] in ['', u'']:  # remove null string
                     scope = scope[:-1]
-
         else:
             scope = None
-        cmd = '(slime::slime-find-keyword "{0}" (lisp:quote {1}) "{2}")'.format(
+
+        # Try to eval from repl_socket to cope with thread special symbols
+        lock = self.euslisp.euslime_connection_lock
+        force_repl = True
+
+        # But use internal_socket when repl is not available
+        if lock.locked():
+            log.warning('Euslisp process is busy! ' +
+                        'Calculating completions on internal process instead')
+            lock = self.euslisp.euslime_internal_connection_lock
+            force_repl = False
+        else:
+            log.debug('Acquiring lock: %s' % lock)
+            lock.acquire()
+
+        cmd = '(slime::slime-find-keyword "{}" (lisp:quote {}) "{}")'.format(
             qstr(start), dumps_lisp(scope), qstr(self.package))
-        # Eval from repl_socket to cope with thread special symbols
-        result = self.euslisp.exec_internal(cmd, force_repl_socket=True)
-        yield EuslispResult(result)
+        try:
+            result = self.euslisp.exec_internal(
+                cmd, force_repl_socket=force_repl)
+            yield EuslispResult(result)
+        finally:
+            if force_repl and lock.locked():
+                lock.release()
 
     def swank_completions_for_character(self, start):
         cmd = '(slime::slime-find-character "{0}")'.format(qstr(start))
@@ -356,11 +459,6 @@ class EuslimeHandler(object):
 
         def check_key(key):
             return key in res_dict and num == res_dict[key]
-        def debug_return(db):
-            msg = db.message.split(self.euslisp.delim)[0]
-            msg = repr(msg.rsplit(' in ', 1)[0])
-            yield [Symbol(':debug-return'), 0, db.level, Symbol('nil')]
-            yield [Symbol(':return'), {'abort': msg}, db.id]
 
         if check_key('QUIT'):
             clear_stack = True
@@ -381,7 +479,7 @@ class EuslimeHandler(object):
             for val in self.maybe_new_prompt():
                 yield val
             for db in reversed(self.debugger):
-                for val in debug_return(db):
+                for val in db.make_debug_return(delim=self.euslisp.delim):
                     yield val
             self.debugger = []
         else:
@@ -390,7 +488,7 @@ class EuslimeHandler(object):
             if not self.debugger:
                 for val in self.maybe_new_prompt():
                     yield val
-            for val in debug_return(deb):
+            for val in deb.make_debug_return(delim=self.euslisp.delim):
                 yield val
             # Pop previous debugger, if any
             if self.debugger:
@@ -428,7 +526,7 @@ class EuslimeHandler(object):
         lock.acquire()
         try:
             for res in self.euslisp.eval(cmd_str):
-                if isinstance(res,list):
+                if isinstance(res, list):
                     yield res
             for msg in messages:
                 yield [Symbol(":write-string"), "; Loaded {}\n".format(msg)]
@@ -442,7 +540,8 @@ class EuslimeHandler(object):
             if lock.locked():
                 lock.release()
             log.info('Aborting evaluation...')
-            # Force print the message, which is by default only displayed on the minibuffer
+            # Force-print the message, which is
+            # by default only displayed on the minibuffer
             if e.message:
                 yield [Symbol(":write-string"),
                        "; Evaluation aborted on {}\n".format(e.message),
@@ -467,22 +566,42 @@ class EuslimeHandler(object):
                              seconds, loadp, filename])
 
     def swank_load_file(self, filename):
+        # Block both streams while loading a file (issue #10)
+        # Maybe there are other alternatives?
         lock = self.euslisp.euslime_connection_lock
         log.debug('Acquiring lock: %s' % lock)
         lock.acquire()
+        internal_lock = self.euslisp.euslime_internal_connection_lock
+        log.debug('Acquiring lock: %s' % internal_lock)
+        internal_lock.acquire()
         yield [Symbol(":write-string"), "Loading file: %s ...\n" % filename]
         try:
-            for r in self.euslisp.eval('(lisp:load "{0}")'.format(qstr(filename))):
-                if isinstance(r,list):
+            cmd = '(slime::load-file-and-tags "{0}")'.format(qstr(filename))
+            output_str = ''
+
+            for r in self.euslisp.eval(cmd):
+                if isinstance(r, list):
+                    # read-string
                     yield r
+                if isinstance(r, str):
+                    output_str += r
+
+            res = loads(output_str, nil=None)
+            if res == Symbol("lisp:nil") or res == Symbol("nil"):
+                res = None
+
             yield [Symbol(":write-string"), "Loaded.\n"]
-            yield EuslispResult(True)
+            yield EuslispResult(res)
             lock.release()
+            internal_lock.release()
         except AbortEvaluation as e:
             if lock.locked():
                 lock.release()
+            if internal_lock.locked():
+                internal_lock.release()
             log.info('Aborting evaluation...')
-            # Force print the message, which is by default only displayed on the minibuffer
+            # Force-print the message, which is
+            # by default only displayed on the minibuffer
             if e.message:
                 yield [Symbol(":write-string"),
                        "; Evaluation aborted on {}\n".format(e.message),
@@ -491,6 +610,8 @@ class EuslimeHandler(object):
         except Exception:
             if lock.locked():
                 lock.release()
+            if internal_lock.locked():
+                internal_lock.release()
             raise
 
     def swank_inspect_current_condition(self):
@@ -504,6 +625,9 @@ class EuslimeHandler(object):
         return
 
     def swank_frame_locals_and_catch_tags(self, *args):
+        # [RET] on sldb Backtrace entry
+        # Display local bindings at the given frame
+        # e.g. (:return (:ok (((:name "NUMBER" :id 0 :value "NIL")) nil)) 7)
         return
 
     def swank_find_definitions_for_emacs(self, keyword):
@@ -546,16 +670,21 @@ class EuslimeHandler(object):
     def swank_repl_clear_repl_variables(self):
         lock = self.euslisp.euslime_connection_lock
         log.debug('Acquiring lock: %s' % lock)
+
+        if lock.locked():
+            log.error('Could not acquire lock: %s' % lock)
+            yield EuslispResult("'Process busy'", response_type='abort')
+            return
+
         lock.acquire()
         try:
             cmd = '(slime::clear-repl-variables)'
             res = self.euslisp.exec_internal(cmd, force_repl_socket=True)
             yield EuslispResult(res)
             lock.release()
-        except Exception:
+        finally:
             if lock.locked():
                 lock.release()
-            raise
 
     def swank_clear_repl_results(self):
         return
@@ -563,16 +692,31 @@ class EuslimeHandler(object):
     def swank_set_package(self, name):
         lock = self.euslisp.euslime_connection_lock
         log.debug('Acquiring lock: %s' % lock)
+
+        if not check_lock(lock, 0.5):
+            log.error('Could not acquire lock: %s' % lock)
+            yield [Symbol(":write-string"),
+                   "; No responses from inferior process\n"]
+            yield EuslispResult(False, response_type='abort')
+            return
+
         lock.acquire()
         try:
             cmd = '(slime::set-package "{0}")'.format(qstr(name))
             res = self.euslisp.exec_internal(cmd, force_repl_socket=True)
             yield EuslispResult(res)
-            lock.release()
-        except Exception:
+        except AbortEvaluation:
+            # the abort signal sometimes comes earlier than the
+            # evaluation result, for example when attaching a gdb
+            # instance and passing a SIGINT.
+            # Retry the evaluation in such cases
+            log.info('AbortEvaluation received during swank_set_package.' +
+                     ' Retrying...')
+            res = self.euslisp.exec_internal(cmd, force_repl_socket=True)
+            yield EuslispResult(res)
+        finally:
             if lock.locked():
                 lock.release()
-            raise
 
     def swank_default_directory(self):
         cmd = '(lisp:pwd)'
